import time

from paasta_tools.deployd.common import PaastaThread


class QueueMetrics(PaastaThread):
    def __init__(self, inbox, bounce_q, cluster, metrics_provider):
        super().__init__()
        self.daemon = True
        self.inbox_q = inbox.inbox_q
        self.inbox = inbox.to_bounce
        self.bounce_q = bounce_q
        self.metrics = metrics_provider
        self.inbox_q_gauge = self.metrics.create_gauge("inbox_queue", paasta_cluster=cluster)
        self.inbox_gauge = self.metrics.create_gauge("inbox", paasta_cluster=cluster)
        self.bounce_q_gauge = self.metrics.create_gauge("bounce_queue", paasta_cluster=cluster)

    def run(self):
        while True:
            self.inbox_q_gauge.set(self.inbox_q.qsize())
            self.inbox_gauge.set(len(self.inbox.keys()))
            self.bounce_q_gauge.set(self.bounce_q.qsize())
            time.sleep(20)



#!/usr/bin/env python
import argparse
import itertools

from paasta_tools.marathon_tools import get_list_of_marathon_clients
from paasta_tools.utils import paasta_print


def parse_args():
    parser = argparse.ArgumentParser(
        description=(
            'Helper script to get the number of in-progress marathon deployments'
        ),
    )
    return parser.parse_args()


def get_deployments():
    clients = get_list_of_marathon_clients()
    return [deployment for deployment in itertools.chain.from_iterable(
        c.list_deployments() for c in clients
    )]


def main():
    parse_args()
    paasta_print(len(get_deployments()))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to wait for deployment
of a docker image to a cluster.instance.
"""
import logging

from paasta_tools.cli.cmds.mark_for_deployment import NoSuchCluster
from paasta_tools.cli.cmds.mark_for_deployment import report_waiting_aborted
from paasta_tools.cli.cmds.mark_for_deployment import wait_for_deployment
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import list_services
from paasta_tools.cli.utils import NoSuchService
from paasta_tools.cli.utils import validate_git_sha
from paasta_tools.cli.utils import validate_given_deploy_groups
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.cli.utils import validate_short_git_sha
from paasta_tools.remote_git import list_remote_refs
from paasta_tools.remote_git import LSRemoteException
from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_git_url
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import TimeoutError


DEFAULT_DEPLOYMENT_TIMEOUT = 3600  # seconds


log = logging.getLogger(__name__)


class GitShaError(Exception):
    pass


class DeployGroupError(Exception):
    pass


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        'wait-for-deployment',
        help='Wait a service to be deployed to deploy_group',
        description=(
            "'paasta wait-for-deployment' waits for a previously marked for "
            "deployment service to be deployed to deploy_group."
        ),
        epilog=(
            "Note: Access and credentials to the Git repo of a service "
            "are required for this command to work."
        ),
    )
    list_parser.add_argument(
        '-u', '--git-url',
        help=(
            'Git url for service. Defaults to the normal git URL for '
            'the service.'
        ),
        default=None,
    )
    list_parser.add_argument(
        '-c', '-k', '--commit',
        help='Git sha to wait for deployment',
        required=True,
        type=validate_short_git_sha,
    )
    list_parser.add_argument(
        '-l', '--deploy-group',
        help='deploy group (e.g. cluster1.canary, cluster2.main).',
        required=True,
    ).completer = lazy_choices_completer(list_deploy_groups)
    list_parser.add_argument(
        '-s', '--service',
        help='Name of the service which you wish to wait for deployment. '
        'Leading "services-" will be stripped.',
        required=True,
    ).completer = lazy_choices_completer(list_services)
    list_parser.add_argument(
        '-t', '--timeout',
        dest="timeout",
        type=int,
        default=DEFAULT_DEPLOYMENT_TIMEOUT,
        help=(
            "Time in seconds to wait for paasta to deploy the service. "
            "If the timeout is exceeded we return 1. "
            "Default is %(default)s seconds."
        ),
    )
    list_parser.add_argument(
        '-d', '--soa-dir',
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    list_parser.add_argument(
        '-v', '--verbose',
        action='count',
        dest="verbose",
        default=0,
        help="Print out more output.",
    )

    list_parser.set_defaults(command=paasta_wait_for_deployment)


def get_latest_marked_sha(git_url, deploy_group):
    """Return the latest marked for deployment git sha or ''"""
    refs = list_remote_refs(git_url)
    last_ref = ''
    for ref in refs:
        if (
            ref.startswith(f'refs/tags/paasta-{deploy_group}-') and
            ref.endswith('-deploy') and
            ref > last_ref
        ):
            last_ref = ref
    return refs[last_ref] if last_ref else ''


def validate_git_sha_is_latest(git_sha, git_url, deploy_group, service):
    """Verify if git_sha is the latest sha marked for deployment.

    Raise exception when the provided git_sha is not the latest
    marked for deployment in 'deploy_group' for 'service'.
    """
    try:
        marked_sha = get_latest_marked_sha(git_url, deploy_group)
    except LSRemoteException as e:
        paasta_print("Error talking to the git server: {}\n"
                     "It is not possible to verify that {} is marked for deployment in {}, "
                     "but I assume that it is marked and will continue waiting.."
                     .format(e, git_sha, deploy_group))
        return
    if marked_sha == '':
        raise GitShaError("ERROR: Nothing is marked for deployment "
                          "in {} for {}"
                          .format(deploy_group, service))
    if git_sha != marked_sha:
        raise GitShaError("ERROR: The latest git SHA marked for "
                          "deployment in {} is {}"
                          .format(deploy_group, marked_sha))


def validate_deploy_group(deploy_group, service, soa_dir):
    """Validate deploy_group.

    Raise exception if the specified deploy group is not used anywhere.
    """
    in_use_deploy_groups = list_deploy_groups(
        service=service,
        soa_dir=soa_dir,
    )
    _, invalid_deploy_groups = \
        validate_given_deploy_groups(in_use_deploy_groups, [deploy_group])

    if len(invalid_deploy_groups) == 1:
        raise DeployGroupError("ERROR: These deploy groups are not currently "
                               "used anywhere: {}.\n"
                               "You probably need one of these in-use deploy "
                               "groups?:\n   {}"
                               .format(
                                   ",".join(invalid_deploy_groups),
                                   ",".join(in_use_deploy_groups),
                               ))


def paasta_wait_for_deployment(args):
    """Wrapping wait_for_deployment"""
    if args.verbose:
        log.setLevel(level=logging.DEBUG)
    else:
        log.setLevel(level=logging.INFO)

    service = args.service
    if service and service.startswith('services-'):
        service = service.split('services-', 1)[1]

    if args.git_url is None:
        args.git_url = get_git_url(service=service, soa_dir=args.soa_dir)

    args.commit = validate_git_sha(sha=args.commit, git_url=args.git_url)

    try:
        validate_service_name(service, soa_dir=args.soa_dir)
        validate_deploy_group(args.deploy_group, service, args.soa_dir)
        validate_git_sha_is_latest(
            args.commit, args.git_url,
            args.deploy_group, service,
        )
    except (GitShaError, DeployGroupError, NoSuchService) as e:
        paasta_print(PaastaColors.red(f'{e}'))
        return 1

    try:
        wait_for_deployment(
            service=service,
            deploy_group=args.deploy_group,
            git_sha=args.commit,
            soa_dir=args.soa_dir,
            timeout=args.timeout,
        )
        _log(
            service=service,
            component='deploy',
            line=("Deployment of {} for {} complete".
                  format(args.commit, args.deploy_group)),
            level='event',
        )

    except (KeyboardInterrupt, TimeoutError, NoSuchCluster):
        report_waiting_aborted(service, args.deploy_group)
        return 1

    return 0



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from paasta_tools.api import client
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_instances
from paasta_tools.cli.utils import list_services
from paasta_tools.utils import list_clusters
from paasta_tools.utils import paasta_print


log = logging.getLogger(__name__)


def add_subparser(subparsers):
    autoscale_parser = subparsers.add_parser(
        'autoscale',
        help="Manually scale a service up and down manually, bypassing the normal autoscaler",
    )

    autoscale_parser.add_argument(
        '-s', '--service',
        help="Service that you want to stop. Like 'example_service'.",
    ).completer = lazy_choices_completer(list_services)
    autoscale_parser.add_argument(
        '-i', '--instance',
        help="Instance of the service that you want to stop. Like 'main' or 'canary'.",
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    autoscale_parser.add_argument(
        '-c', '--cluster',
        help="The PaaSTA cluster that has the service instance you want to stop. Like 'norcal-prod'.",
        required=True,
    ).completer = lazy_choices_completer(list_clusters)
    autoscale_parser.add_argument(
        '--set',
        help="Set the number to scale to. Must be an Int.",
        type=int,
    )
    autoscale_parser.set_defaults(command=paasta_autoscale)


def paasta_autoscale(args):
    log.setLevel(logging.DEBUG)
    service = figure_out_service_name(args)
    api = client.get_paasta_api_client(cluster=args.cluster, http_res=True)
    if not api:
        paasta_print('Could not connect to paasta api. Maybe you misspelled the cluster?')
        return 1

    if args.set is None:
        log.debug("Getting the current autoscaler count...")
        res, http = api.autoscaler.get_autoscaler_count(service=service, instance=args.instance).result()
    else:
        log.debug(f"Setting desired instances to {args.set}.")
        body = {'desired_instances': int(args.set)}
        res, http = api.autoscaler.update_autoscaler_count(
            service=service, instance=args.instance, json_body=body,
        ).result()

    log.debug(f"Res: {res} Http: {http}")
    print(res["desired_instances"])
    return 0



import sys
from collections import defaultdict

from paasta_tools.utils import get_docker_client


def main():
    docker_client = get_docker_client()
    ip_to_containers = defaultdict(list)

    for container in docker_client.containers():
        networks = container['NetworkSettings']['Networks']
        if 'bridge' in networks:
            ip = networks['bridge']['IPAddress']
            if ip:
                ip_to_containers[ip].append(container)

    output = []
    for ip, containers in ip_to_containers.items():
        if len(containers) > 1:
            output.append(f'{ip} shared by the following containers:')
            for container in containers:
                output.append('    Image: {}'.format(container['Image']))
                output.append('        ID: {}'.format(container['Id']))
                output.append('        State: {}'.format(container['State']))
                output.append('        Status: {}'.format(container['Status']))
            output.append('')

    if output:
        print('CRITICAL - There are multiple Docker containers assigned to the same IP.')
        print('There should only be one per IP. Choose one to keep and try stopping the others.')
        print('\n'.join(output))
        return 2
    else:
        print('OK - No Docker containers sharing an IP on this host.')
        return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
import argparse

from boto3 import session


def get_client(endpoint):
    s = session.Session(
        region_name='foo',
        aws_access_key_id='foo',
        aws_secret_access_key='bar',
    )
    client = s.client(
        service_name='dynamodb',
        endpoint_url=endpoint,
    )
    return client


def create_table(client, table_name):
    return client.create_table(
        TableName=table_name,
        KeySchema=[
            {
                'AttributeName': 'task_id',
                'KeyType': 'HASH',
            },
            {
                'AttributeName': 'timestamp',
                'KeyType': 'RANGE',
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'task_id',
                'AttributeType': 'S',
            },
            {
                'AttributeName': 'timestamp',
                'AttributeType': 'N',
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 123,
            'WriteCapacityUnits': 123,
        },
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Create a dummy dynamodb table')
    parser.add_argument('endpoint', type=str, help='the dynamodb endpoint')
    parser.add_argument('table_name', type=str, help='the name of the table to create')
    args = parser.parse_args()
    create_table(get_client(args.endpoint), args.table_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import errno
import json


DEFAULTS = {
    "debug": "false",
    "log_file": None,
    "log_level": "warning",
    "master": "localhost:5050",
    "max_workers": 5,
    "scheme": "http",
    "response_timeout": 5,
}


def load_mesos_config(config_path, profile="default"):
    on_disk = {}

    try:
        with open(config_path, 'rt') as f:
            on_disk = json.load(f)[profile]
    except ValueError as e:
        raise ValueError(
            'Invalid JSON: {} in {}'.format(str(e), config_path),
        )
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise

    config = copy.deepcopy(DEFAULTS)
    config.update(on_disk)
    return config



#!/usr/bin/env python
# Copyright 2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Marathon Dashboard
"""
from pyramid.view import view_config

from paasta_tools.api import settings
from paasta_tools.marathon_dashboard import create_marathon_dashboard


@view_config(route_name='marathon_dashboard', request_method='GET', renderer='json')
def marathon_dashboard(request):
    print("marathon_dashboard view")
    return create_marathon_dashboard(
        cluster=settings.cluster,
        soa_dir=settings.soa_dir,
        marathon_clients=settings.marathon_clients,
        system_paasta_config=settings.system_paasta_config,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fnmatch
import itertools
import json
import logging
import os
import re
from typing import List
from urllib.parse import urljoin
from urllib.parse import urlparse

import aiohttp
from kazoo.handlers.threading import KazooTimeoutError
from kazoo.retry import KazooRetry
from mypy_extensions import TypedDict
from retry import retry

from . import exceptions
from . import framework
from . import log
from . import mesos_file
from . import slave
from . import task
from . import util
from . import zookeeper
from paasta_tools.async_utils import async_ttl_cache
from paasta_tools.utils import get_user_agent

ZOOKEEPER_TIMEOUT = 1

INVALID_PATH = "{0} does not have a valid path. Did you forget /mesos?"

MISSING_MASTER = """unable to connect to a master at {0}.

Try running `mesos config master zk://localhost:2181/mesos`. See the README for
more examples."""

MULTIPLE_SLAVES = "There are multiple slaves with that id. Please choose one: "

logger = logging.getLogger(__name__)


MesosState = TypedDict(
    'MesosState',
    {
        'slaves': List,
    },
)


class MesosMaster:

    def __init__(self, config):
        self.config = config

    def __str__(self):
        return "<master: {}>".format(self.key())

    def key(self):
        return self.config["master"]

    @util.CachedProperty(ttl=5)
    def host(self):
        return "{}://{}".format(
            self.config["scheme"],
            self.resolve(self.config["master"]),
        )

    @util.CachedProperty(ttl=5)
    def cache_host(self):
        host_url = urlparse(self.host)
        replaced = host_url._replace(netloc=host_url.hostname + ':5055')
        return replaced.geturl()

    async def _request(
        self,
        url: str,
        method: str='GET',
        cached: bool=False,
        **kwargs,
    ) -> aiohttp.ClientResponse:
        headers = {'User-Agent': get_user_agent()}

        if cached and self.config.get("use_mesos_cache", False):
            # TODO: fall back to original host if this fails?
            host = self.cache_host
        else:
            host = self.host

        try:
            async with aiohttp.ClientSession(
                conn_timeout=self.config["response_timeout"],
                read_timeout=self.config["response_timeout"],
            ) as session:
                async with session.request(
                    method=method,
                    url=urljoin(host, url),
                    headers=headers,
                    **kwargs,
                ) as resp:
                    # if nobody awaits resp.text() or resp.json() before we exit the session context manager, then the
                    # http connection gets closed before we read the response; then later calls to resp.text/json will
                    # fail.
                    await resp.text()
                    return resp

        except aiohttp.client_exceptions.ClientConnectionError:
            raise exceptions.MasterNotAvailableException(
                MISSING_MASTER.format(host),
            )
        except aiohttp.client_exceptions.TooManyRedirects:
            raise exceptions.MasterTemporarilyNotAvailableException(
                (
                    "Unable to connect to master at %s, likely due to "
                    "an ongoing leader election"
                ) % host,
            )

    async def fetch(self, url, **kwargs):
        return await self._request(url, **kwargs)

    async def post(self, url, **kwargs):
        return await self._request(url, method='POST', **kwargs)

    def _file_resolver(self, cfg):
        return self.resolve(open(cfg[6:], "r+").read().strip())

    @retry(KazooTimeoutError, tries=5, delay=0.5, logger=logger)
    def _zookeeper_resolver(self, cfg):
        hosts, path = cfg[5:].split("/", 1)
        path = "/" + path

        retry = KazooRetry(max_tries=10)
        with zookeeper.client(
            hosts=hosts,
            read_only=True,
            connection_retry=retry,
            command_retry=retry,
        ) as zk:
            def master_id(key):
                return int(key.split("_")[-1])

            def get_masters():
                return [x for x in zk.get_children(path)
                        if re.search("\d+", x)]

            leader = sorted(get_masters(), key=lambda x: master_id(x))

            if len(leader) == 0:
                raise exceptions.MasterNotAvailableException(
                    f"cannot find any masters at {cfg}",
                )
            data, stat = zk.get(os.path.join(path, leader[0]))

            if not data:
                exceptions.MasterNotAvailableException(
                    "Cannot retrieve valid MasterInfo data from ZooKeeper",
                )
            else:
                data = data.decode('utf8')

            try:
                parsed = json.loads(data)
                if parsed and "address" in parsed:
                    ip = parsed["address"].get("ip")
                    port = parsed["address"].get("port")
                    if ip and port:
                        return f"{ip}:{port}"
            except ValueError as parse_error:
                log.debug(
                    "[WARN] No JSON content, probably connecting to older "
                    "Mesos version. Reason: {}".format(parse_error),
                )
                raise exceptions.MasterNotAvailableException(
                    "Failed to parse mesos master ip from ZK",
                )

    @log.duration
    def resolve(self, cfg):
        """Resolve the URL to the mesos master.

        The value of cfg should be one of:
            - host:port
            - zk://host1:port1,host2:port2/path
            - zk://username:password@host1:port1/path
            - file:///path/to/file (where file contains one of the above)
        """
        if cfg.startswith("zk:"):
            return self._zookeeper_resolver(cfg)
        elif cfg.startswith("file:"):
            return self._file_resolver(cfg)
        else:
            return cfg

    @async_ttl_cache(ttl=15)
    async def state(self) -> MesosState:
        return await (await self.fetch("/master/state.json", cached=True)).json()

    async def state_summary(self) -> MesosState:
        return await (await self.fetch("/master/state-summary")).json()

    @async_ttl_cache(ttl=0)
    async def slave(self, fltr):
        lst = await self.slaves(fltr)

        log.debug(f"master.slave({fltr})")

        if len(lst) == 0:
            raise exceptions.SlaveDoesNotExist(
                f"Slave {fltr} no longer exists.",
            )

        elif len(lst) > 1:
            raise exceptions.MultipleSlavesForIDError(
                "Multiple slaves matching filter {}. {}".format(
                    fltr,
                    ",".join([slave.id for slave in lst]),
                ),
            )

        return lst[0]

    async def slaves(self, fltr=""):
        return [
            slave.MesosSlave(self.config, x)
            for x in (await self.state())['slaves']
            if fltr == x['id']
        ]

    async def _task_list(self, active_only=False):
        keys = ["tasks"]
        if not active_only:
            keys.append("completed_tasks")
        return itertools.chain(
            *[util.merge(x, *keys) for x in await self._framework_list(active_only)],
        )

    async def task(self, fltr):
        lst = await self.tasks(fltr)

        if len(lst) == 0:
            raise exceptions.TaskNotFoundException(
                "Cannot find a task with filter %s" % fltr,
            )

        elif len(lst) > 1:
            raise exceptions.MultipleTasksForIDError(
                "Multiple tasks matching filter {}. {}".format(
                    fltr,
                    ",".join([task.id for task in lst]),
                ),
            )
        return lst[0]

    async def orphan_tasks(self):
        return (await self.state())["orphan_tasks"]

    # XXX - need to filter on task state as well as id
    async def tasks(self, fltr="", active_only=False):
        return [
            task.Task(self, x)
            for x in await self._task_list(active_only)
            if fltr in x['id'] or fnmatch.fnmatch(x['id'], fltr)
        ]

    async def framework(self, fwid):
        return list(filter(
            lambda x: x.id == fwid,
            await self.frameworks(),
        ))[0]

    async def _framework_list(self, active_only=False):
        keys = ["frameworks"]
        if not active_only:
            keys.append("completed_frameworks")
        return util.merge(await self._frameworks(), *keys)

    @async_ttl_cache(ttl=15)
    async def _frameworks(self):
        return await (await self.fetch("/master/frameworks", cached=True)).json()

    async def frameworks(self, active_only=False):
        return [framework.Framework(f)
                for f in await self._framework_list(active_only)]

    async def teardown(self, framework_id):
        return await self.post(
            "/master/teardown", data="frameworkId=%s" % framework_id,
        )

    async def metrics_snapshot(self):
        return await (await self.fetch("/metrics/snapshot")).json()

    @property  # type: ignore
    @util.memoize
    def log(self):
        return mesos_file.File(self, path="/master/log")



import logging
import socket
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import service_configuration_lib
from kazoo.exceptions import NoNodeError
from mypy_extensions import TypedDict

from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import InstanceConfigDict
from paasta_tools.utils import InvalidInstanceConfig
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import ZookeeperPool

DEFAULT_CONTAINER_PORT = 8888

log = logging.getLogger(__name__)
logging.getLogger('marathon').setLevel(logging.WARNING)

AUTOSCALING_ZK_ROOT = '/autoscaling'
ZK_PAUSE_AUTOSCALE_PATH = '/autoscaling/paused'


class LongRunningServiceConfigDict(InstanceConfigDict, total=False):
    drain_method: str
    container_port: int
    drain_method_params: Dict
    healthcheck_cmd: str
    healthcheck_grace_period_seconds: float
    healthcheck_interval_seconds: float
    healthcheck_max_consecutive_failures: int
    healthcheck_mode: str
    healthcheck_timeout_seconds: float
    healthcheck_uri: str
    instances: int
    max_instances: int
    min_instances: int
    nerve_ns: str
    registrations: List[str]
    bounce_priority: int


# Defined here to avoid import cycles -- this gets used in bounce_lib and subclassed in marathon_tools.
BounceMethodConfigDict = TypedDict('BounceMethodConfigDict', {"instances": int})


class ServiceNamespaceConfig(dict):

    def get_mode(self) -> str:
        """Get the mode that the service runs in and check that we support it.
        If the mode is not specified, we check whether the service uses smartstack
        in order to determine the appropriate default value. If proxy_port is specified
        in the config, the service uses smartstack, and we can thus safely assume its mode is http.
        If the mode is not defined and the service does not use smartstack, we set the mode to None.
        """
        mode = self.get('mode', None)
        if mode is None:
            if not self.is_in_smartstack():
                return None
            else:
                return 'http'
        elif mode in ['http', 'tcp', 'https']:
            return mode
        else:
            raise InvalidSmartstackMode("Unknown mode: %s" % mode)

    def get_healthcheck_uri(self) -> str:
        return self.get('healthcheck_uri', '/status')

    def get_discover(self) -> str:
        return self.get('discover', 'region')

    def is_in_smartstack(self) -> bool:
        if self.get('proxy_port') is not None:
            return True
        else:
            return False


class LongRunningServiceConfig(InstanceConfig):
    config_dict: LongRunningServiceConfigDict

    def __init__(
        self, service: str, cluster: str, instance: str, config_dict: LongRunningServiceConfigDict,
        branch_dict: Optional[BranchDictV2], soa_dir: str=DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )

    def get_drain_method(self, service_namespace_config: ServiceNamespaceConfig) -> str:
        """Get the drain method specified in the service's marathon configuration.

        :param service_config: The service instance's configuration dictionary
        :returns: The drain method specified in the config, or 'noop' if not specified"""
        default = 'noop'
        # Default to hacheck draining if the service is in smartstack
        if service_namespace_config.is_in_smartstack():
            default = 'hacheck'
        return self.config_dict.get('drain_method', default)

    def get_drain_method_params(self, service_namespace_config: ServiceNamespaceConfig) -> Dict:
        """Get the drain method parameters specified in the service's marathon configuration.

        :param service_config: The service instance's configuration dictionary
        :returns: The drain_method_params dictionary specified in the config, or {} if not specified"""
        default: Dict = {}
        if service_namespace_config.is_in_smartstack():
            default = {'delay': 60}
        return self.config_dict.get('drain_method_params', default)

    # FIXME(jlynch|2016-08-02, PAASTA-4964): DEPRECATE nerve_ns and remove it
    def get_nerve_namespace(self) -> str:
        return decompose_job_id(self.get_registrations()[0])[1]

    def get_registrations(self) -> List[str]:
        registrations = self.config_dict.get('registrations', [])
        for registration in registrations:
            try:
                decompose_job_id(registration)
            except InvalidJobNameError:
                log.error(
                    'Provided registration {} for service '
                    '{} is invalid'.format(registration, self.service),
                )

        # Backwards compatibility with nerve_ns
        # FIXME(jlynch|2016-08-02, PAASTA-4964): DEPRECATE nerve_ns and remove it
        if not registrations and 'nerve_ns' in self.config_dict:
            registrations.append(
                compose_job_id(self.service, self.config_dict['nerve_ns']),
            )

        return registrations or [compose_job_id(self.service, self.instance)]

    def get_healthcheck_uri(self, service_namespace_config: ServiceNamespaceConfig) -> str:
        return self.config_dict.get('healthcheck_uri', service_namespace_config.get_healthcheck_uri())

    def get_healthcheck_cmd(self) -> str:
        cmd = self.config_dict.get('healthcheck_cmd', None)
        if cmd is None:
            raise InvalidInstanceConfig("healthcheck mode 'cmd' requires a healthcheck_cmd to run")
        else:
            return cmd

    def get_healthcheck_grace_period_seconds(self) -> float:
        """How long Marathon should give a service to come up before counting failed healthchecks."""
        return self.config_dict.get('healthcheck_grace_period_seconds', 60)

    def get_healthcheck_interval_seconds(self) -> float:
        return self.config_dict.get('healthcheck_interval_seconds', 10)

    def get_healthcheck_timeout_seconds(self) -> float:
        return self.config_dict.get('healthcheck_timeout_seconds', 10)

    def get_healthcheck_max_consecutive_failures(self) -> int:
        return self.config_dict.get('healthcheck_max_consecutive_failures', 30)

    def get_healthcheck_mode(self, service_namespace_config: ServiceNamespaceConfig) -> str:
        mode = self.config_dict.get('healthcheck_mode', None)
        if mode is None:
            mode = service_namespace_config.get_mode()
        elif mode not in ['http', 'https', 'tcp', 'cmd', None]:
            raise InvalidHealthcheckMode("Unknown mode: %s" % mode)
        return mode

    def get_bounce_priority(self) -> int:
        """Gives a priority to each service instance which deployd will use to prioritise services.
        Higher numbers are higher priority. This affects the order in which deployd workers pick
        instances from the bounce queue.

        NB: we multiply by -1 here because *internally* lower numbers are higher priority.
        """
        return self.config_dict.get('bounce_priority', 0) * -1

    def get_instances(self) -> int:
        """Gets the number of instances for a service, ignoring whether the user has requested
        the service to be started or stopped"""
        if self.get_max_instances() is not None:
            try:
                zk_instances = get_instances_from_zookeeper(
                    service=self.service,
                    instance=self.instance,
                )
                log.debug("Got %d instances out of zookeeper" % zk_instances)
            except NoNodeError:
                log.debug("No zookeeper data, returning max_instances (%d)" % self.get_max_instances())
                return self.get_max_instances()
            else:
                limited_instances = self.limit_instance_count(zk_instances)
                return limited_instances
        else:
            instances = self.config_dict.get('instances', 1)
            log.debug("Autoscaling not enabled, returning %d instances" % instances)
            return instances

    def get_min_instances(self) -> int:
        return self.config_dict.get('min_instances', 1)

    def get_max_instances(self) -> int:
        return self.config_dict.get('max_instances', None)

    def get_desired_instances(self) -> int:
        """Get the number of instances specified in zookeeper or the service's marathon configuration.
        If the number of instances in zookeeper is less than min_instances, returns min_instances.
        If the number of instances in zookeeper is greater than max_instances, returns max_instances.

        Defaults to 0 if not specified in the config.

        :returns: The number of instances specified in the config, 0 if not
                  specified or if desired_state is not 'start'.
                  """
        if self.get_desired_state() == 'start':
            return self.get_instances()
        else:
            log.debug("Instance is set to stop. Returning '0' instances")
            return 0

    def limit_instance_count(self, instances: int) -> int:
        """
        Returns param instances if it is between min_instances and max_instances.
        Returns max_instances if instances > max_instances
        Returns min_instances if instances < min_instances
        """
        return max(
            self.get_min_instances(),
            min(self.get_max_instances(), instances),
        )

    def get_container_port(self) -> int:
        return self.config_dict.get('container_port', DEFAULT_CONTAINER_PORT)


class InvalidHealthcheckMode(Exception):
    pass


def get_healthcheck_for_instance(
    service: str,
    instance: str,
    service_manifest: LongRunningServiceConfig,
    random_port: int,
    soa_dir: str=DEFAULT_SOA_DIR,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns healthcheck for a given service instance in the form of a tuple (mode, healthcheck_command)
    or (None, None) if no healthcheck
    """
    namespace = service_manifest.get_nerve_namespace()
    smartstack_config = load_service_namespace_config(
        service=service,
        namespace=namespace,
        soa_dir=soa_dir,
    )
    mode = service_manifest.get_healthcheck_mode(smartstack_config)
    hostname = socket.getfqdn()

    if mode == "http" or mode == "https":
        path = service_manifest.get_healthcheck_uri(smartstack_config)
        healthcheck_command = '%s://%s:%d%s' % (mode, hostname, random_port, path)
    elif mode == "tcp":
        healthcheck_command = '%s://%s:%d' % (mode, hostname, random_port)
    elif mode == 'cmd':
        healthcheck_command = service_manifest.get_healthcheck_cmd()
    else:
        mode = None
        healthcheck_command = None
    return (mode, healthcheck_command)


def load_service_namespace_config(service: str, namespace: str, soa_dir: str=DEFAULT_SOA_DIR) -> ServiceNamespaceConfig:
    """Attempt to read the configuration for a service's namespace in a more strict fashion.

    Retrieves the following keys:

    - proxy_port: the proxy port defined for the given namespace
    - healthcheck_mode: the mode for the healthcheck (http or tcp)
    - healthcheck_port: An alternate port to use for health checking
    - healthcheck_uri: URI target for healthchecking
    - healthcheck_timeout_s: healthcheck timeout in seconds
    - healthcheck_body_expect: an expected string in healthcheck response body
    - updown_timeout_s: updown_service timeout in seconds
    - timeout_connect_ms: proxy frontend timeout in milliseconds
    - timeout_server_ms: proxy server backend timeout in milliseconds
    - timeout_client_ms: proxy server client timeout in milliseconds
    - retries: the number of retries on a proxy backend
    - mode: the mode the service is run in (http or tcp)
    - routes: a list of tuples of (source, destination)
    - discover: the scope at which to discover services e.g. 'habitat'
    - advertise: a list of scopes to advertise services at e.g. ['habitat', 'region']
    - extra_advertise: a list of tuples of (source, destination)
      e.g. [('region:dc6-prod', 'region:useast1-prod')]
    - extra_healthcheck_headers: a dict of HTTP headers that must
      be supplied when health checking. E.g. { 'Host': 'example.com' }

    :param service: The service name
    :param namespace: The namespace to read
    :param soa_dir: The SOA config directory to read from
    :returns: A dict of the above keys, if they were defined
    """

    service_config = service_configuration_lib.read_service_configuration(
        service_name=service, soa_dir=soa_dir,
    )
    smartstack_config = service_config.get('smartstack', {})
    namespace_config_from_file = smartstack_config.get(namespace, {})

    service_namespace_config = ServiceNamespaceConfig()
    # We can't really use .get, as we don't want the key to be in the returned
    # dict at all if it doesn't exist in the config file.
    # We also can't just copy the whole dict, as we only care about some keys
    # and there's other things that appear in the smartstack section in
    # several cases.
    key_whitelist = {
        'healthcheck_mode',
        'healthcheck_uri',
        'healthcheck_port',
        'healthcheck_timeout_s',
        'healthcheck_body_expect',
        'updown_timeout_s',
        'proxy_port',
        'timeout_connect_ms',
        'timeout_server_ms',
        'timeout_client_ms',
        'retries',
        'mode',
        'discover',
        'advertise',
        'extra_healthcheck_headers',
    }

    for key, value in namespace_config_from_file.items():
        if key in key_whitelist:
            service_namespace_config[key] = value

    # Other code in paasta_tools checks 'mode' after the config file
    # is loaded, so this ensures that it is set to the appropriate default
    # if not otherwise specified, even if appropriate default is None.
    service_namespace_config['mode'] = service_namespace_config.get_mode()

    if 'routes' in namespace_config_from_file:
        service_namespace_config['routes'] = [(route['source'], dest)
                                              for route in namespace_config_from_file['routes']
                                              for dest in route['destinations']]

    if 'extra_advertise' in namespace_config_from_file:
        service_namespace_config['extra_advertise'] = [
            (src, dst)
            for src in namespace_config_from_file['extra_advertise']
            for dst in namespace_config_from_file['extra_advertise'][src]
        ]

    return service_namespace_config


class InvalidSmartstackMode(Exception):
    pass


def get_instances_from_zookeeper(service: str, instance: str) -> int:
    with ZookeeperPool() as zookeeper_client:
        (instances, _) = zookeeper_client.get('%s/instances' % compose_autoscaling_zookeeper_root(service, instance))
        return int(instances)


def compose_autoscaling_zookeeper_root(service: str, instance: str) -> str:
    return f'{AUTOSCALING_ZK_ROOT}/{service}/{instance}'


def set_instances_for_marathon_service(
    service: str,
    instance: str,
    instance_count: int,
    soa_dir: str=DEFAULT_SOA_DIR,
) -> None:
    zookeeper_path = '%s/instances' % compose_autoscaling_zookeeper_root(service, instance)
    with ZookeeperPool() as zookeeper_client:
        zookeeper_client.ensure_path(zookeeper_path)
        zookeeper_client.set(zookeeper_path, str(instance_count).encode('utf8'))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from urllib.parse import urljoin

import aiohttp

from . import exceptions
from . import mesos_file
from . import util
from paasta_tools.async_utils import async_ttl_cache
from paasta_tools.utils import get_user_agent


class MesosSlave:

    def __init__(self, config, items):
        self.config = config
        self.__items = items

    def __getitem__(self, name):
        return self.__items[name]

    def __str__(self):
        return self.key()

    def key(self):
        return self["pid"].split('@')[-1]

    @property
    def host(self):
        return "{}://{}:{}".format(
            self.config["scheme"],
            self["hostname"],
            self["pid"].split(":")[-1],
        )

    async def fetch(self, url, **kwargs) -> aiohttp.ClientResponse:
        headers = {'User-Agent': get_user_agent()}
        async with aiohttp.ClientSession(
            conn_timeout=self.config["response_timeout"],
            read_timeout=self.config["response_timeout"],
        ) as session:
            try:
                async with session.get(
                    urljoin(self.host, url),
                    headers=headers,
                    **kwargs,
                ) as response:
                    await response.text()
                    return response
            except aiohttp.ClientConnectionError:
                raise exceptions.SlaveDoesNotExist(
                    f"Unable to connect to the slave at {self.host}",
                )

    @async_ttl_cache(ttl=5)
    async def state(self):
        return await (await self.fetch("/slave(1)/state.json")).json()

    async def frameworks(self):
        return util.merge(await self.state(), "frameworks", "completed_frameworks")

    async def task_executor(self, task_id):
        for fw in await self.frameworks():
            for exc in util.merge(fw, "executors", "completed_executors"):
                if task_id in list(map(
                        lambda x: x["id"],
                        util.merge(
                            exc, "completed_tasks", "tasks", "queued_tasks",
                        ),
                )):
                    return exc
        raise exceptions.MissingExecutor("No executor has a task by that id")

    async def file_list(self, path):
        # The sandbox does not exist on the slave.
        if path == "":
            return []

        resp = self.fetch("/files/browse.json", params={"path": path})
        if resp.status_code == 404:
            return []
        return await resp.json()

    def file(self, task, path):
        return mesos_file.File(self, task, path)

    @async_ttl_cache(ttl=1)
    async def stats(self):
        return await (await self.fetch("/monitor/statistics.json")).json()

    def executor_stats(self, _id):
        return list(filter(lambda x: x["executor_id"]))

    async def task_stats(self, _id):
        stats = list(filter(
            lambda x: x["executor_id"] == _id,
            await self.stats(),
        ))

        # Tasks that are not yet in a RUNNING state have no stats.
        if len(stats) == 0:
            return {}
        else:
            return stats[0]["statistics"]

    @property  # type: ignore
    @util.memoize
    def log(self):
        return mesos_file.File(self, path="/slave/log")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import itertools
import time


def merge(obj, *keys):
    return itertools.chain(*[obj[k] for k in keys])


class CachedProperty:

    def __init__(self, ttl=300):
        self.ttl = ttl

    def __call__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        return self

    def __get__(self, inst, owner):
        try:
            value, last_update = inst._cache[self.__name__]
            if self.ttl > 0 and time.time() - last_update > self.ttl:
                raise AttributeError
        except (KeyError, AttributeError):
            value = self.fget(inst)
            try:
                cache = inst._cache
            except AttributeError:
                cache = inst._cache = {}
            cache[self.__name__] = (value, time.time())
        return value


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


def humanize_bytes(b):
    abbrevs = (
        (1 << 30, 'GB'),
        (1 << 20, 'MB'),
        (1 << 10, 'kB'),
        (1, 'B'),
    )
    for factor, suffix in abbrevs:
        if b >= factor:
            break
    return '%.*f %s' % (2, b / float(factor), suffix)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.autoscaling.pause_service_autoscaler import delete_service_autoscale_pause_time
from paasta_tools.autoscaling.pause_service_autoscaler import get_service_autoscale_pause_time
from paasta_tools.autoscaling.pause_service_autoscaler import update_service_autoscale_pause_time
from paasta_tools.utils import paasta_print

MAX_PAUSE_DURATION = 320


def add_subparser(subparsers):
    status_parser = subparsers.add_parser(
        'pause_service_autoscaler',
        help="Pause the service autoscaler for an entire cluster",
        description=(
            "'paasta pause_service_autoscaler is used to pause the paasta service autoscaler "
            "for an entire paasta cluster. "
        ),
    )
    status_parser.add_argument(
        '-c', '--cluster',
        dest="cluster",
        help='which cluster to pause autoscaling in. ie. norcal-prod',
    )
    status_parser.add_argument(
        '-d', '--pause-duration',
        default=120,
        dest="duration",
        type=int,
        help="How long to pause the autoscaler for, defaults to 120 minutes",
    )
    status_parser.add_argument(
        '-f', '--force',
        help='Force pause for longer than max duration',
        action='store_true',
        dest='force',
        default=False,
    )
    status_parser.add_argument(
        '-i', '--info',
        help='Print when the autoscaler is paused until',
        action='store_true',
        dest='info',
        default=False,
    )
    status_parser.add_argument(
        '-r', '--resume',
        help='Resume autoscaling (unpause) in a cluster',
        action='store_true',
        dest='resume',
        default=False,
    )

    status_parser.set_defaults(command=paasta_pause_service_autoscaler)


def paasta_pause_service_autoscaler(args):
    """With a given cluster and duration, pauses the paasta service autoscaler
       in that cluster for duration minutes"""
    if args.duration > MAX_PAUSE_DURATION:
        if not args.force:
            paasta_print('Specified duration: {d} longer than max: {m}'.format(
                d=args.duration,
                m=MAX_PAUSE_DURATION,
            ))
            paasta_print('If you are really sure, run again with --force')
            return 3

    if args.info:
        return_code = get_service_autoscale_pause_time(args.cluster)
    elif args.resume:
        return_code = delete_service_autoscale_pause_time(args.cluster)
    else:
        minutes = args.duration
        return_code = update_service_autoscale_pause_time(args.cluster, minutes)

    return return_code



import staticconf

CLUSTERMAN_YAML_FILE_PATH = '/nail/srv/configs/clusterman.yaml'
CLUSTERMAN_METRICS_YAML_FILE_PATH = '/nail/srv/configs/clusterman_metrics.yaml'


def get_clusterman_metrics():
    try:
        import clusterman_metrics
        clusterman_yaml = CLUSTERMAN_YAML_FILE_PATH
        staticconf.YamlConfiguration(CLUSTERMAN_METRICS_YAML_FILE_PATH, namespace='clusterman_metrics')
    except ImportError:
        # our cluster autoscaler is not currently open source, sorry!
        clusterman_metrics = None
        clusterman_yaml = None

    return clusterman_metrics, clusterman_yaml



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys

from paasta_tools.smartstack_tools import are_services_up_on_ip_port
from paasta_tools.utils import load_system_paasta_config

system_paasta_config = load_system_paasta_config()
synapse_port = system_paasta_config.get_synapse_port()
synapse_host = '169.254.255.254'
synapse_haproxy_url_format = system_paasta_config.get_synapse_haproxy_url_format()
host_ip = os.environ['PAASTA_POD_IP']
services = sys.argv[1:]

if are_services_up_on_ip_port(
    synapse_host=synapse_host,
    synapse_port=synapse_port,
    synapse_haproxy_url_format=synapse_haproxy_url_format,
    services=services,
    host_ip=host_ip,
    host_port=8888,
):
    sys.exit(0)
else:
    sys.exit(1)



#!/usr/bin/env python
import logging

import yelp_meteorite

from paasta_tools.cli.utils import get_instance_config
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def emit_metrics_for_type(instance_type):
    cluster = load_system_paasta_config().get_cluster()
    instances = get_services_for_cluster(
        cluster=cluster,
        instance_type=instance_type,
    )

    for service, instance in instances:
        service_instance_config = get_instance_config(
            service=service,
            instance=instance,
            cluster=cluster,
        )
        dimensions = {
            'paasta_service': service_instance_config.service,
            'paasta_cluster': service_instance_config.cluster,
            'paasta_instance': service_instance_config.instance,
        }

        log.info(f"Emitting paasta.service.* with dimensions {dimensions}")
        gauge = yelp_meteorite.create_gauge('paasta.service.cpus', dimensions)
        gauge.set(service_instance_config.get_cpus())
        gauge = yelp_meteorite.create_gauge('paasta.service.mem', dimensions)
        gauge.set(service_instance_config.get_mem())
        gauge = yelp_meteorite.create_gauge('paasta.service.disk', dimensions)
        gauge.set(service_instance_config.get_disk())
        if hasattr(service_instance_config, 'get_instances'):
            if service_instance_config.get_max_instances() is None:
                gauge = yelp_meteorite.create_gauge('paasta.service.instances', dimensions)
                gauge.set(service_instance_config.get_instances())


def main():
    logging.basicConfig(level=logging.INFO)
    for thing in ["marathon", "chronos", "adhoc"]:
        emit_metrics_for_type(thing)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import copy
from typing import Any
from typing import List
from typing import Optional

import service_configuration_lib
from mypy_extensions import TypedDict

from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.long_running_service_tools import LongRunningServiceConfig
from paasta_tools.long_running_service_tools import LongRunningServiceConfigDict
from paasta_tools.long_running_service_tools import ServiceNamespaceConfig
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import Constraint  # noqa, imported for typing.
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DockerParameter
from paasta_tools.utils import get_code_sha_from_dockerurl
from paasta_tools.utils import get_config_hash
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import paasta_print
from paasta_tools.utils import SystemPaastaConfig


MESOS_TASK_SPACER = '.'


VolumeInfo = TypedDict(
    'VolumeInfo',
    {
        'container_path': str,
        'host_path': str,
        'mode': str,
    },
)

_Docker_PortMapping = TypedDict(
    '_Docker_PortMapping',
    {
        'host_port': int,
        'container_port': int,
        'protocol': str,
    },
)

DockerInfo = TypedDict(
    'DockerInfo',
    {
        'image': str,
        'network': str,
        'port_mappings': List[_Docker_PortMapping],
        'parameters': List[DockerParameter],
    },
)

ContainerInfo = TypedDict(
    'ContainerInfo',
    {
        'type': str,
        'docker': DockerInfo,
        'volumes': List[VolumeInfo],
    },
)


_CommandInfo_URI = TypedDict(
    '_CommandInfo_URI',
    {
        'value': str,
        'executable': bool,
        'extract': bool,
        'cache': bool,
        'output_file': str,
    },
    total=False,
)

CommandInfo = TypedDict(
    'CommandInfo',
    {
        'uris': List[_CommandInfo_URI],
        'environment': Any,
        'shell': bool,
        'value': str,
        'arguments': List[str],
        'user': str,
    },
    total=False,
)

Value_Scalar = TypedDict(
    'Value_Scalar',
    {
        'value': float,
    },
)

Value_Range = TypedDict(
    'Value_Range',
    {
        'begin': int,
        'end': int,
    },
)

Value_Ranges = TypedDict(
    'Value_Ranges',
    {
        'range': List[Value_Range],
    },
)

Resource = TypedDict(
    'Resource',
    {
        'name': str,
        'type': str,
        'scalar': Value_Scalar,
        'ranges': Value_Ranges,
    },
    total=False,
)


TaskID = TypedDict(
    'TaskID',
    {
        'value': str,
    },
)

SlaveID = TypedDict(
    'SlaveID',
    {
        'value': str,
    },
)


class TaskInfoBase(TypedDict):
    name: str
    task_id: TaskID
    agent_id: SlaveID
    resources: List[Resource]


class TaskInfo(TaskInfoBase):
    container: ContainerInfo
    command: CommandInfo


class NativeServiceConfigDict(LongRunningServiceConfigDict):
    pass


class NativeServiceConfig(LongRunningServiceConfig):
    config_dict: NativeServiceConfigDict
    config_filename_prefix = 'paasta_native'

    def __init__(
        self,
        service: str,
        instance: str,
        cluster: str,
        config_dict: NativeServiceConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str,
        service_namespace_config: Optional[ServiceNamespaceConfig]=None,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )
        # service_namespace_config may be omitted/set to None at first, then set
        # after initializing. e.g. we do this in load_paasta_native_job_config
        # so we can call get_nerve_namespace() to figure out what SNC to read.
        # It may also be set to None if this service is not in nerve.
        if service_namespace_config is not None:
            self.service_namespace_config = service_namespace_config
        else:
            self.service_namespace_config = ServiceNamespaceConfig()

    def task_name(self, base_task: TaskInfo) -> str:
        code_sha = get_code_sha_from_dockerurl(
            base_task['container']['docker']['image'],
        )

        filled_in_task = copy.deepcopy(base_task)
        filled_in_task['name'] = ''
        filled_in_task['task_id'] = {'value': ''}
        filled_in_task['agent_id'] = {'value': ''}

        config_hash = get_config_hash(
            filled_in_task,
            force_bounce=self.get_force_bounce(),
        )

        return compose_job_id(
            self.service,
            self.instance,
            git_hash=code_sha,
            config_hash=config_hash,
            spacer=MESOS_TASK_SPACER,
        )

    def base_task(self, system_paasta_config: SystemPaastaConfig, portMappings=True) -> TaskInfo:
        """Return a TaskInfo Dict with all the fields corresponding to the
        configuration filled in.

        Does not include task.agent_id or a task.id; those need to be
        computed separately.
        """
        docker_volumes = self.get_volumes(
            system_volumes=system_paasta_config.get_volumes(),
        )
        task: TaskInfo = {
            'name': '',
            'task_id': {'value': ''},
            'agent_id': {'value': ''},
            'container': {
                'type': 'DOCKER',
                'docker': {
                    'image': self.get_docker_url(),
                    'parameters': [
                        {'key': param['key'], 'value': param['value']}
                        for param in self.format_docker_parameters()
                    ],
                    'network': self.get_mesos_network_mode(),
                    'port_mappings': [],
                },
                'volumes': [
                    {
                        'container_path': volume['containerPath'],
                        'host_path': volume['hostPath'],
                        'mode': volume['mode'].upper(),
                    }
                    for volume in docker_volumes
                ],
            },
            'command': {
                'value': self.get_cmd(),
                'uris': [
                    {
                        'value': system_paasta_config.get_dockercfg_location(),
                        'extract': False,
                    },
                ],
            },
            'resources': [
                {
                    'name': 'cpus',
                    'type': 'SCALAR',
                    'scalar': {'value': self.get_cpus()},
                },
                {
                    'name': 'mem',
                    'type': 'SCALAR',
                    'scalar': {'value': self.get_mem()},
                },
            ],
        }

        if portMappings:
            task['container']['docker']['port_mappings'] = [
                {
                    'container_port': self.get_container_port(),
                    # filled by tasks_and_state_for_offer()
                    'host_port': 0,
                    'protocol': 'tcp',
                },
            ]

            task['resources'].append(
                {
                    'name': 'ports',
                    'type': 'RANGES',
                    'ranges': {
                        # filled by tasks_and_state_for_offer
                        'range': [{'begin': 0, 'end': 0}],
                    },
                },
            )

        task['name'] = self.task_name(task)

        return task

    def get_mesos_network_mode(self) -> str:
        return self.get_net().upper()


def load_paasta_native_job_config(
    service,
    instance,
    cluster,
    load_deployments=True,
    soa_dir=DEFAULT_SOA_DIR,
    instance_type='paasta_native',
    config_overrides=None,
) -> NativeServiceConfig:
    service_paasta_native_jobs = read_service_config(
        service=service,
        instance=instance,
        instance_type=instance_type,
        cluster=cluster,
        soa_dir=soa_dir,
    )
    branch_dict: Optional[BranchDictV2] = None
    instance_config_dict = service_paasta_native_jobs[instance].copy()
    instance_config_dict.update(config_overrides or {})
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = NativeServiceConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=instance_config_dict,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    service_config = NativeServiceConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=instance_config_dict,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )

    service_namespace_config = load_service_namespace_config(
        service=service,
        namespace=service_config.get_nerve_namespace(),
        soa_dir=soa_dir,
    )
    service_config.service_namespace_config = service_namespace_config

    return service_config


def read_service_config(service, instance, instance_type, cluster, soa_dir=DEFAULT_SOA_DIR):
    conf_file = f'{instance_type}-{cluster}'
    full_path = f'{soa_dir}/{service}/{conf_file}.yaml'
    paasta_print("Reading paasta-remote configuration file: %s" % full_path)

    config = service_configuration_lib.read_extra_service_information(
        service,
        conf_file,
        soa_dir=soa_dir,
    )

    if instance not in config:
        raise UnknownNativeServiceError(
            'No job named "{}" in config file {}: \n{}'.format(
                instance, full_path, open(full_path).read(),
            ),
        )

    return config


class UnknownNativeServiceError(Exception):
    pass



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.cli.utils import execute_paasta_serviceinit_on_remote_master
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_instances
from paasta_tools.cli.utils import list_services
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors


def add_subparser(subparsers):
    status_parser = subparsers.add_parser(
        'emergency-start',
        help="Kicks off a chronos job run. Not implemented for Marathon instances.",
        description=(
            "Chronos Jobs: Forces a job to run outside of its normal schedule.\n"
            "Marathon Apps: Not implemented.\n"
        ),
    )
    status_parser.add_argument(
        '-s', '--service',
        help="Service that you want to start. Like 'example_service'.",
    ).completer = lazy_choices_completer(list_services)
    status_parser.add_argument(
        '-i', '--instance',
        help="Instance of the service that you want to start. Like 'main' or 'canary'.",
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    status_parser.add_argument(
        '-c', '--cluster',
        help="The PaaSTA cluster that has the service instance you want to start. Like 'norcal-prod'.",
        required=True,
    ).completer = lazy_choices_completer(list_clusters)
    status_parser.add_argument(
        '-d', '--soa-dir',
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    status_parser.set_defaults(command=paasta_emergency_start)


def paasta_emergency_start(args):
    """Performs an emergency start on a given service instance on a given cluster

    All it does for Chronos jobs is send the latest version of the job config to Chronos and run it immediately.
    """
    system_paasta_config = load_system_paasta_config()
    service = figure_out_service_name(args, soa_dir=args.soa_dir)
    paasta_print("Performing an emergency start on %s..." % compose_job_id(service, args.instance))
    return_code, output = execute_paasta_serviceinit_on_remote_master(
        subcommand='start',
        cluster=args.cluster,
        service=service,
        instances=args.instance,
        system_paasta_config=system_paasta_config,
    )
    paasta_print("%s" % "\n".join(paasta_emergency_start.__doc__.splitlines()[-8:]))
    paasta_print("Output: %s" % PaastaColors.grey(output))
    paasta_print("Run this command to see the status:")
    paasta_print(f"paasta status --service {service} --clusters {args.cluster}")

    return return_code



#!/usr/bin/env python
import argparse
import sys
import time
from time import sleep

from paasta_tools import mesos_tools
from paasta_tools.frameworks.native_scheduler import create_driver
from paasta_tools.frameworks.native_scheduler import get_paasta_native_jobs_for_cluster
from paasta_tools.frameworks.native_scheduler import load_paasta_native_job_config
from paasta_tools.frameworks.native_scheduler import NativeScheduler
from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PaastaNotConfiguredError


def parse_args(argv):
    parser = argparse.ArgumentParser(description='Runs native paasta mesos scheduler.')
    parser.add_argument('-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR", default=DEFAULT_SOA_DIR)
    parser.add_argument('--stay-alive-seconds', dest="stay_alive_seconds", type=int, default=300)
    parser.add_argument('--periodic-interval', dest="periodic_interval", type=int, default=30)
    parser.add_argument('--staging-timeout', dest="staging_timeout", type=float, default=60)
    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    system_paasta_config = load_system_paasta_config()
    cluster = system_paasta_config.get_cluster()

    drivers = []
    schedulers = []
    for service, instance in get_paasta_native_jobs_for_cluster(cluster=cluster, soa_dir=args.soa_dir):
        scheduler = NativeScheduler(
            service_name=service,
            instance_name=instance,
            cluster=cluster,
            staging_timeout=args.staging_timeout,
            system_paasta_config=system_paasta_config,
            soa_dir=args.soa_dir,
        )
        schedulers.append(scheduler)

        driver = create_driver(
            framework_name="paasta_native %s" % compose_job_id(service, instance),
            scheduler=scheduler,
            system_paasta_config=system_paasta_config,
        )
        driver.start()
        drivers.append(driver)

    end_time = time.time() + args.stay_alive_seconds
    while time.time() < end_time:
        sleep(args.periodic_interval)
        for scheduler, driver in zip(schedulers, drivers):
            scheduler.periodic(driver)

    return schedulers


def get_app_id_and_task_uuid_from_executor_id(executor_id):
    """Parse the paasta_native executor ID and return the (app id, task uuid)"""
    return executor_id.rsplit('.', 1)


def parse_service_instance_from_executor_id(task_id):
    app_id, task_uuid = get_app_id_and_task_uuid_from_executor_id(task_id)
    (srv_name, srv_instance, _, __) = decompose_job_id(app_id)
    return srv_name, srv_instance


def paasta_native_services_running_here(hostname=None, framework_id=None):
    """See what paasta_native services are being run by a mesos-slave on this host.
    :returns: A list of triples of (service, instance, port)

    :param hostname: query the mesos slave on this hostname.
    :param framework_id: If specified, return info only for tasks belonging to this framework id.
    """

    def framework_filter(fw):
        return fw['name'].startswith('paasta_native ') and (framework_id is None or fw['id'] == framework_id)

    return mesos_tools.mesos_services_running_here(
        framework_filter=framework_filter,
        parse_service_instance_from_executor_id=parse_service_instance_from_executor_id,
        hostname=hostname,
    )


def get_paasta_native_services_running_here_for_nerve(cluster, soa_dir, hostname=None):
    if not cluster:
        try:
            system_paasta_config = load_system_paasta_config()
        # In the cases where there is *no* cluster or in the case
        # where there isn't a Paasta configuration file at *all*, then
        # there must be no native services running here, so we catch
        # these custom exceptions and return [].
        except (PaastaNotConfiguredError):
            return []
        cluster = system_paasta_config.get_cluster()
        if not system_paasta_config.get_register_native_services():
            return []
    # When a cluster is defined in mesos, let's iterate through paasta_native services
    paasta_native_services = paasta_native_services_running_here(hostname=hostname)
    nerve_list = []
    for name, instance, port in paasta_native_services:
        try:
            registrations = read_all_registrations_for_service_instance(
                name, instance, cluster, soa_dir,
            )
            for registration in registrations:
                reg_service, reg_namespace, _, __ = decompose_job_id(registration)
                nerve_dict = load_service_namespace_config(
                    service=reg_service, namespace=reg_namespace, soa_dir=soa_dir,
                )
                if not nerve_dict.is_in_smartstack():
                    continue
                nerve_dict['port'] = port
                nerve_list.append((registration, nerve_dict))
        except KeyError:
            continue  # SOA configs got deleted for this app, it'll get cleaned up
    return nerve_list


def read_all_registrations_for_service_instance(service, instance, cluster=None, soa_dir=DEFAULT_SOA_DIR):
    """Retrieve all registrations as fully specified name.instance pairs
    for a particular service instance.

    For example, the 'main' paasta instance of the 'test' service may register
    in the 'test.main' namespace as well as the 'other_svc.main' namespace.

    If one is not defined in the config file, returns a list containing
    name.instance instead.
    """
    if not cluster:
        cluster = load_system_paasta_config().get_cluster()

    job_config = load_paasta_native_job_config(
        service=service,
        instance=instance,
        cluster=cluster,
        load_deployments=False,
        soa_dir=soa_dir,
    )
    return job_config.get_registrations()


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import sys
import traceback
from socket import getfqdn
from socket import gethostbyname
from socket import gethostname

from paasta_tools import mesos_maintenance
from paasta_tools import utils
from paasta_tools.marathon_tools import get_expected_instance_count_for_namespace
from paasta_tools.marathon_tools import marathon_services_running_here
from paasta_tools.marathon_tools import read_registration_for_service_instance
from paasta_tools.smartstack_tools import backend_is_up
from paasta_tools.smartstack_tools import get_backends
from paasta_tools.smartstack_tools import get_replication_for_services
from paasta_tools.smartstack_tools import ip_port_hostname_from_svname
from paasta_tools.smartstack_tools import load_smartstack_info_for_service
from paasta_tools.utils import paasta_print

log = logging.getLogger(__name__)


def parse_args():
    """Parses the command line arguments passed to this script"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-d', '--duration',
        type=mesos_maintenance.parse_timedelta,
        default='1h',
        help="Duration of the maintenance window. Any pytimeparse unit is supported.",
    )
    parser.add_argument(
        '-s', '--start',
        type=mesos_maintenance.parse_datetime,
        default=str(mesos_maintenance.now()),
        help="Time to start the maintenance window. Defaults to now.",
    )
    parser.add_argument(
        'action',
        choices=[
            'cluster_status',
            'down',
            'drain',
            'is_host_down',
            'is_host_drained',
            'is_host_draining',
            'is_hosts_past_maintenance_end',
            'is_hosts_past_maintenance_start',
            'is_safe_to_drain',
            'is_safe_to_kill',
            'schedule',
            'status',
            'undrain',
            'up',
        ],
        help="Action to perform on the specified hosts",
    )
    parser.add_argument(
        'hostname',
        nargs='*',
        default=[getfqdn()],
        help='Hostname(s) of machine(s) to start draining. '
        'You can specify <hostname>|<ip> to avoid querying DNS to determine the corresponding IP.',
    )
    parser.add_argument(
        '-v', '--verbose',
        action='count',
        dest="verbose",
        default=0,
        help="Print out more output.",
    )
    return parser.parse_args()


def is_safe_to_kill(hostname):
    """Checks if a host has drained or reached its maintenance window
    :param hostname: hostname to check
    :returns: True or False
    """
    return mesos_maintenance.is_host_drained(hostname) or \
        mesos_maintenance.is_host_past_maintenance_start(hostname)


def is_hostname_local(hostname):
    return hostname == 'localhost' or \
        hostname == getfqdn() or hostname == gethostname()


def is_safe_to_drain(hostname):
    """Checks if a host has healthy tasks running locally that have low
    replication in other places
    :param hostname: hostname to check
    :returns: True or False
    """
    if not is_hostname_local(hostname):
        paasta_print("Due to the way is_safe_to_drain is implemented, it can only work on localhost.")
        return False
    return not are_local_tasks_in_danger()


def is_healthy_in_haproxy(local_port, backends):
    local_ip = gethostbyname(gethostname())
    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend['svname'])
        if ip == local_ip and port == local_port:
            if backend_is_up(backend):
                log.debug("Found a healthy local backend: %s" % backend)
                return True
            else:
                log.debug("Found a unhealthy local backend: %s" % backend)
                return False
    log.debug("Couldn't find any haproxy backend listening on %s" % local_port)
    return False


def synapse_replication_is_low(service, instance, system_paasta_config, local_backends):
    crit_threshold = 80
    reg_svc, reg_namespace, _, __ = utils.decompose_job_id(
        read_registration_for_service_instance(
            service=service, instance=instance,
        ),
    )
    # We only actually care about the replication of where we're registering
    service, namespace = reg_svc, reg_namespace

    smartstack_replication_info = load_smartstack_info_for_service(
        service=service,
        namespace=namespace,
        blacklist=[],
        system_paasta_config=system_paasta_config,
    )
    expected_count = get_expected_instance_count_for_namespace(service=service, namespace=namespace)
    expected_count_per_location = int(expected_count / len(smartstack_replication_info))

    synapse_name = utils.compose_job_id(service, namespace)
    local_replication = get_replication_for_services(
        synapse_host=system_paasta_config.get_default_synapse_host(),
        synapse_port=system_paasta_config.get_synapse_port(),
        synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
        services=[synapse_name],
    )
    num_available = local_replication.get(synapse_name, 0)
    under_replicated, ratio = utils.is_under_replicated(
        num_available, expected_count_per_location, crit_threshold,
    )
    log.info('Service %s.%s has %d out of %d expected instances' % (
        service, instance, num_available, expected_count_per_location,
    ))
    return under_replicated


def are_local_tasks_in_danger():
    try:
        system_paasta_config = utils.load_system_paasta_config()
        local_services = marathon_services_running_here()
        local_backends = get_backends(
            service=None,
            synapse_host=system_paasta_config.get_default_synapse_host(),
            synapse_port=system_paasta_config.get_synapse_port(),
            synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
        )
        for service, instance, port in local_services:
            log.info(f"Inspecting {service}.{instance} on {port}")
            if is_healthy_in_haproxy(port, local_backends) and \
               synapse_replication_is_low(service, instance, system_paasta_config, local_backends=local_backends):
                log.warning("{}.{} on port {} is healthy but the service is in danger!".format(
                    service, instance, port,
                ))
                return True
        return False
    except Exception:
        log.warning(traceback.format_exc())
        return False


def paasta_maintenance():
    """Manipulate the maintenance state of a PaaSTA host.
    :returns: None
    """
    args = parse_args()

    if args.verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    action = args.action
    hostnames = args.hostname

    if action != 'status' and not hostnames:
        paasta_print("You must specify one or more hostnames")
        return

    start = args.start
    duration = args.duration

    ret = "Done"
    if action == 'drain':
        mesos_maintenance.drain(hostnames, start, duration)
    elif action == 'undrain':
        mesos_maintenance.undrain(hostnames)
    elif action == 'down':
        mesos_maintenance.down(hostnames)
    elif action == 'up':
        mesos_maintenance.up(hostnames)
    elif action == 'status':
        ret = mesos_maintenance.friendly_status()
    elif action == 'cluster_status':
        ret = mesos_maintenance.status()
    elif action == 'schedule':
        ret = mesos_maintenance.schedule()
    elif action == 'is_safe_to_drain':
        ret = is_safe_to_drain(hostnames[0])
    elif action == 'is_safe_to_kill':
        ret = is_safe_to_kill(hostnames[0])
    elif action == 'is_host_drained':
        ret = mesos_maintenance.is_host_drained(hostnames[0])
    elif action == 'is_host_down':
        ret = mesos_maintenance.is_host_down(hostnames[0])
    elif action == 'is_host_draining':
        ret = mesos_maintenance.is_host_draining(hostnames[0])
    elif action == 'is_host_past_maintenance_start':
        ret = mesos_maintenance.is_host_past_maintenance_start(hostnames[0])
    elif action == 'is_host_past_maintenance_end':
        ret = mesos_maintenance.is_host_past_maintenance_end(hostnames[0])
    else:
        raise NotImplementedError("Action: '%s' is not implemented." % action)
    paasta_print(ret)
    return ret


if __name__ == '__main__':
    if paasta_maintenance():
        sys.exit(0)
    sys.exit(1)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob

from pkg_resources import yield_lines
from setuptools import find_packages
from setuptools import setup

from paasta_tools import __version__


def get_install_requires():
    with open('requirements-minimal.txt', 'r') as f:
        minimal_reqs = list(yield_lines(f.read()))

    return minimal_reqs


setup(
    name='paasta-tools',
    version=__version__,
    provides=["paasta_tools"],
    author='Kyle Anderson',
    author_email='kwa@yelp.com',
    description='Tools for Yelps SOA infrastructure',
    packages=find_packages(exclude=("tests*", "scripts*")),
    include_package_data=True,
    install_requires=get_install_requires(),
    scripts=[
        'paasta_tools/am_i_mesos_leader.py',
        'paasta_tools/autoscale_all_services.py',
        'paasta_tools/check_marathon_services_replication.py',
        'paasta_tools/check_oom_events.py',
        'paasta_tools/cleanup_marathon_jobs.py',
        'paasta_tools/paasta_deploy_chronos_jobs',
        'paasta_tools/deploy_marathon_services',
        'paasta_tools/paasta_deploy_tron_jobs',
        'paasta_tools/generate_all_deployments',
        'paasta_tools/generate_deployments_for_service.py',
        'paasta_tools/generate_services_file.py',
        'paasta_tools/generate_services_yaml.py',
        'paasta_tools/get_mesos_leader.py',
        'paasta_tools/list_marathon_service_instances.py',
        'paasta_tools/marathon_dashboard.py',
        'paasta_tools/monitoring/check_capacity.py',
        'paasta_tools/monitoring/check_chronos_has_jobs.py',
        'paasta_tools/monitoring/check_marathon_has_apps.py',
        'paasta_tools/monitoring/check_mesos_active_frameworks.py',
        'paasta_tools/monitoring/check_mesos_duplicate_frameworks.py',
        'paasta_tools/monitoring/check_mesos_quorum.py',
        'paasta_tools/monitoring/check_mesos_outdated_tasks.py',
        'paasta_tools/monitoring/check_synapse_replication.py',
        'paasta_tools/monitoring/kill_orphaned_docker_containers.py',
        'paasta_tools/cli/paasta_tabcomplete.sh',
        'paasta_tools/paasta_cluster_boost.py',
        'paasta_tools/paasta_execute_docker_command.py',
        'paasta_tools/paasta_maintenance.py',
        'paasta_tools/paasta_metastatus.py',
        'paasta_tools/paasta_remote_run.py',
        'paasta_tools/paasta_serviceinit.py',
        'paasta_tools/setup_kubernetes_job.py',
        'paasta_tools/setup_marathon_job.py',
        'paasta_tools/synapse_srv_namespaces_fact.py',
    ] + glob.glob('paasta_tools/contrib/*'),
    entry_points={
        'console_scripts': [
            'paasta=paasta_tools.cli.cli:main',
            'paasta-api=paasta_tools.api.api:main',
            'paasta-deployd=paasta_tools.deployd.master:main',
            'paasta_autoscale_cluster=paasta_tools.autoscale_cluster:main',
            'paasta_cleanup_chronos_jobs=paasta_tools.cleanup_chronos_jobs:main',
            'paasta_cleanup_tron_namespaces=paasta_tools.cleanup_tron_namespaces:main',
            'paasta_check_chronos_jobs=paasta_tools.check_chronos_jobs:main',
            'paasta_list_chronos_jobs=paasta_tools.list_chronos_jobs:main',
            'paasta_list_kubernetes_service_instances=paasta_tools.list_kubernetes_service_instances:main',
            'paasta_setup_chronos_job=paasta_tools.setup_chronos_job:main',
            'paasta_chronos_rerun=paasta_tools.chronos_rerun:main',
            'paasta_list_tron_namespaces=paasta_tools.list_tron_namespaces:main',
            'paasta_setup_tron_namespace=paasta_tools.setup_tron_namespace:main',
            'paasta_cleanup_maintenance=paasta_tools.cleanup_maintenance:main',
            'paasta_docker_wrapper=paasta_tools.docker_wrapper:main',
            'paasta_firewall_update=paasta_tools.firewall_update:main',
            'paasta_firewall_logging=paasta_tools.firewall_logging:main',
            'paasta_oom_logger=paasta_tools.oom_logger:main',
            'paasta_broadcast_log=paasta_tools.marathon_tools:broadcast_log_all_services_running_here_from_stdin',
        ],
        'paste.app_factory': [
            'paasta-api-config=paasta_tools.api.api:make_app',
        ],
    },
)



import asyncio
import functools
import time
from typing import AsyncIterable
from typing import Awaitable
from typing import Callable
from typing import Dict  # noqa: imported for typing
from typing import List
from typing import TypeVar


T = TypeVar('T')


def async_ttl_cache(
    ttl: float=300,
) -> Callable[
    [Callable[..., Awaitable[T]]],  # wrapped
    Callable[..., Awaitable[T]],  # inner
]:
    _cache: Dict = {}  # Should be Dict[Any, T] but that doesn't work.

    def outer(wrapped):
        @functools.wraps(wrapped)
        async def inner(*args, **kwargs):
            key = functools._make_key(args, kwargs, typed=False)
            try:
                future, last_update = _cache[key]
                if ttl > 0 and time.time() - last_update > ttl:
                    raise KeyError
            except KeyError:
                future = asyncio.ensure_future(wrapped(*args, **kwargs))
                # set the timestamp to +infinity so that we always wait on the in-flight request.
                _cache[key] = (future, float('Inf'))
            value = await future
            _cache[key] = (future, time.time())
            return value
        return inner
    return outer


async def aiter_to_list(
    aiter: AsyncIterable[T],
) -> List[T]:
    return [x async for x in aiter]


def async_timeout(
    seconds: int=10,
) -> Callable[
    [Callable[..., Awaitable[T]]],  # wrapped
    Callable[..., Awaitable[T]],  # inner
]:
    def outer(wrapped):
        @functools.wraps(wrapped)
        async def inner(*args, **kwargs):
            return await asyncio.wait_for(wrapped(*args, **kwargs), timeout=seconds)
        return inner
    return outer
